//
//  clipsave  Saves the image on the clipboard to a PNG file
//
//  how to compile
//    gcc -O2 clipsave.c -o clipsave -lpng
//
#define VERSION "20241104a"

#include <stdlib.h>
#include <stdio.h>
#include <unistd.h>
#include <string.h>
#include <libgen.h>
#include <getopt.h>
#include <png.h>
#include <windows.h>

struct PNG_HANDLE{
  png_structp    pp;
  png_infop      info;
  FILE*          fp;
  unsigned char* lbuf;
  int            width;
  int            bpp;
};

struct PNG_HANDLE* save_png_preproc(char* file, int width, int height, int bpp)
{
  struct PNG_HANDLE* png_handle ;

  if( bpp!=24 && bpp!=32 ){
    return NULL;
  }

  png_handle = malloc(sizeof(struct PNG_HANDLE));
  if( png_handle==NULL ){;
    return NULL;
  }

  if( strcmp(file, "-")==0 ){
    png_handle->fp = stdout;
  }else{
    png_handle->fp = fopen(file, "wb");
    if( png_handle->fp==NULL ){
      free(png_handle);
      return NULL;
    }
  }

  png_handle->pp   = png_create_write_struct(PNG_LIBPNG_VER_STRING, NULL, NULL, NULL);
  png_handle->info = png_create_info_struct(png_handle->pp);

  png_handle->bpp  =  bpp;
  png_handle->lbuf = malloc(width * png_handle->bpp/8);
  if( png_handle->lbuf==NULL ){
    png_destroy_write_struct(&png_handle->pp, &png_handle->info);
    fclose(png_handle->fp);
    free(png_handle);
    return NULL;
  }
  png_handle->width=width ;

  png_init_io(png_handle->pp, png_handle->fp);
  png_set_compression_level(png_handle->pp, 9);
  png_set_IHDR(png_handle->pp, png_handle->info,
               width, height, 8,
               bpp==32 ? PNG_COLOR_TYPE_RGB_ALPHA : PNG_COLOR_TYPE_RGB,
               PNG_INTERLACE_NONE,
               PNG_COMPRESSION_TYPE_DEFAULT, PNG_FILTER_TYPE_DEFAULT) ;

  png_set_sRGB(png_handle->pp, png_handle->info, PNG_sRGB_INTENT_SATURATION);
  png_write_info(png_handle->pp, png_handle->info);

  return png_handle;
}

int save_png_writeline(struct PNG_HANDLE* png_handle, unsigned int* img)
{
  png_bytepp*  pixel_buf[1];
  png_bytepp** pixels;
  int res=-1;

  pixels = &pixel_buf[0];
  switch( png_handle->bpp ){
  case 24 :
    for( int i=0 ; i<png_handle->width ; i++ ){
      png_handle->lbuf[i*3+0] = (img[i]    ) & 0xff;
      png_handle->lbuf[i*3+1] = (img[i]>> 8) & 0xff;
      png_handle->lbuf[i*3+2] = (img[i]>>16) & 0xff;
    }
    pixel_buf[0] = (png_bytepp*)&png_handle->lbuf[0];
    png_write_rows(png_handle->pp, (png_bytepp)pixels, 1);
    res = 0;
    break ;
  case 32 :
    for( int i=0 ; i<png_handle->width ; i++ ){
      png_handle->lbuf[i*4+0] = (img[i]    ) & 0xff;
      png_handle->lbuf[i*4+1] = (img[i]>> 8) & 0xff;
      png_handle->lbuf[i*4+2] = (img[i]>>16) & 0xff;
      png_handle->lbuf[i*4+3] = (img[i]>>24) & 0xff;
    }
    pixel_buf[0] = (png_bytepp*)&png_handle->lbuf[0] ;
    png_write_rows(png_handle->pp, (png_bytepp)pixels, 1);
    res = 0;
    break;
  default :
    break;
  }
  return res;
}

int save_png_postproc(struct PNG_HANDLE* png_handle)
{
  png_write_end(png_handle->pp, png_handle->info);
  png_destroy_write_struct(&png_handle->pp, NULL);
  fclose(png_handle->fp);
  free(png_handle->lbuf);
  free(png_handle);
  png_handle = NULL;

  return 0;
}

const struct option longopts[] = {
  {   "quiet", no_argument,    0, 'q' },
  { "version", no_argument,    0, 'v' },
  {    "help", no_argument,    0, 'h' },
  {      NULL,           0, NULL,   0 }
};
const char* optstr = "qvh";

void print_help(char* progname)
{
  printf("Usage: %s [OPTIONs] FILE\n", progname);
  printf("FILE is the PNG file name to save.\n");
  printf("With no FILE, or when FILE is -, write standard output.\n");
  printf("\n");
  printf("OPTIONs:\n");
  printf("  -q, --quiet   : Does not output error messages.\n");
  printf("  -v, --version : Print version.\n");
  printf("  -h, --help    : Print this help.\n");
}

int main(int argc, char** argv)
{
  char* progname = basename(argv[0]);
  int   res = 1;
  int   quiet = 0;

  int opt;
  int longindex = 0;
  while( (opt = getopt_long(argc, argv, optstr, longopts, &longindex)) != -1 ){
    switch (opt) {
    case 'q':
      quiet = 1;
      break;
    case 'v':
      printf("%s  version %s", progname, VERSION);
      return 0;
      break;
    case 'h':
      print_help(progname);
      return 0;
      break;
    default:
      return res;
      break;
    }
  }

  char* fname;
  if( optind>=argc ){
    fname = "-";
  }else{
    fname = argv[optind];
  }

  OpenClipboard(NULL);

  HGLOBAL hglb;
  hglb = GetClipboardData(CF_DIB);
  if( hglb!=NULL ){
    unsigned int* bufimg=NULL;
    char* dibpt=(char*)(GlobalLock(hglb));
    BITMAPINFOHEADER* bi = (BITMAPINFOHEADER*)(dibpt);
    do{
      bufimg = malloc(bi->biWidth * sizeof(unsigned int));
      if( bufimg==NULL ){
        if( quiet==0 ){
          fprintf(stderr, "Error: Failed allocate buffer.\n");
        }
        break;
      }

      if( bi->biBitCount==24 ){
        unsigned char* srcimg;
        srcimg = (unsigned char*)(dibpt + bi->biSize);
        struct PNG_HANDLE* png_handle;
        png_handle = save_png_preproc(fname,
                                      bi->biWidth,
                                      bi->biHeight,
                                      24);
        if( png_handle==NULL ){
          if( quiet==0 ){
            fprintf(stderr, "Error: Failed save_png_preproc.\n");
          }
          break;
        }
        int width = (((bi->biWidth*3)/4) + ((bi->biWidth*3)%4==0?0:1))*4;
        for( int y=bi->biHeight-1; y>=0; y-- ){
          for( int x=0; x<bi->biWidth; x++ ){
            bufimg[x] = srcimg[(y*width)+x*3+2]    |
                        srcimg[(y*width)+x*3+1]<<8 |
                        srcimg[(y*width)+x*3+0]<<16;
          }
          save_png_writeline(png_handle, bufimg);
        }
        save_png_postproc(png_handle);
        res = 0;

      }else if( bi->biBitCount==32 ){
        unsigned char* srcimg;
        srcimg = (unsigned char*)(dibpt+bi->biSize+(4*3));
        struct PNG_HANDLE* png_handle;
        png_handle = save_png_preproc(fname,
                                      bi->biWidth,
                                      bi->biHeight,
                                      32);
        if( png_handle==NULL ){
          if( quiet==0 ){
            fprintf(stderr, "Error: Failed save_png_preproc.\n");
          }
          break;
        }
        int width = bi->biWidth*4;
        for( int y=bi->biHeight-1; y>=0; y-- ){
          for( int x=0; x<bi->biWidth; x++ ){
            bufimg[x] = srcimg[(y*width)+x*4+3]<<24|
                        srcimg[(y*width)+x*4+2]    |
                        srcimg[(y*width)+x*4+1]<<8 |
                        srcimg[(y*width)+x*4+0]<<16;
          }
          save_png_writeline(png_handle, bufimg);
        }
        save_png_postproc(png_handle);
        res = 0;

      }else{
        if( quiet==0 ){
          fprintf(stderr, "Error: Unsupported bitcount.\n");
        }
      }

    }while(0);
    GlobalUnlock(hglb);
    if( bufimg!=NULL ){
      free(bufimg);
      bufimg = NULL;
    }
  }else{
    if( quiet==0 ){
      fprintf(stderr, "Error: Failed GetClipData.\n");
    }
  }

  CloseClipboard();

  return res;
}
